#ifndef __GRAPH_COMMON_H
#define __GRAPH_COMMON_H

#define N_FILE_ENTRIES        251
#define N_FN_ENTRIES         53
#define N_LINE_ENTRIES         37
#define N_INSTR_ENTRIES         793

#define MAX_x86_INSTR_SIZE    16    // According to ia32 sw dev manual vol 2
#define MIN_LINE_SIZE         16
#define FILE_LEN              256
#define FN_LEN                256

#define INVALID_LINE -1
#define INVALID_UPORT -1
#define VGM_BYTE_INVALID   0xFF
#define VGM_WORD_INVALID   0xFFFFFFFF

#define OPEN_FILE_MAX 10
#define IO_BUF_MAX 64000
#define MAX_USE_PORT 7



#define INPUT_RELEVANT 0xfffe

//#define CONSIDER_POTENTIAL

//#define INTER_CD
//#define TRACK_ARGV
//#define TRACK_INPUT

#define VALUE_WANTED

#define BYTE_GRANUITY

//#define CERTIFY
#define TIMESTAMPED
//#define MAPPING

#define MAX_MAPPING_PORT 3
#define MAX_LINK 20 
#define MAX_MAPPING_ENTRY 200

#ifdef MAPPING
typedef struct _val_link val_link;
struct _val_link {
   //Short cnt;
   //Short alloced;
   Int cnt;
   Int uport;
   //Int vals[MAX_MAPPING_PORT];
   Int * expanded[MAX_MAPPING_PORT];
};

typedef struct _val_mapping val_mapping;
struct _val_mapping {
   Int size;
   UInt * indices;
   val_link ** links;
};
#endif

typedef struct _fnCC fnCC;
struct _fnCC {
   Char*   fn; 
   fnCC*   next;
   //lineCC* lines[N_LINE_ENTRIES];
}; 

typedef struct _fileCC fileCC;
struct _fileCC {
   Char*   file;
   fileCC* next;
   fnCC*   fns[N_FN_ENTRIES];
}; 
typedef struct _instr_info instr_info;
   
typedef struct _fun_entry fun_entry;
struct _fun_entry {
   fun_entry * next;
   //Addr orig_addr;
   fnCC * fn; //use fn as the key
   
   fileCC * file;
   //fnCC * fn;
   Addr orig_addr;

   Int orig_sp_offset;
   Int slack;           //the slack between EBP and ESP because push after setting EBP=ESP
   Int alloc_lv;
   //Int used_lv;
   Int cd_available;
   instr_info * first_bb;
};
   

typedef struct _traced_file traced_file;
struct _traced_file {
   Char*   file;
   Char * is_bracket;
   traced_file* next;
}; 
typedef struct _val_stream val_stream;
   

struct _val_stream {
   Int allocated;
   Int used;
   UInt * iis;
   UInt * insts;
   //UInt * visited;   //only used in slicing
};


//storing the potential edges
typedef struct _potential_dep potential_dep;

struct _potential_dep {
   potential_dep * next;
   instr_info * pred_ii;
   Int pred_inst;
};

typedef struct _forward_edge forward_edge;

struct _forward_edge {
   forward_edge * next;
   UInt * ii;
   Int inst;
   UInt value;
};

//This is for relevant slicing. For loading in the potential CD for loads
typedef struct _potential_cd potential_cd;
struct _potential_cd {
     instr_info * load_ii;
     Int pcd_cnt;
     instr_info ** p_pred_bbs;
     instr_info ** p_branches;
};


struct _instr_info {
   instr_info * next;
   Addr    orig_addr;    //key
/*
   UChar   instr_size;
   UChar   data_size;
*/

   //Addr     bb_addr;
   //the BBs are created when the control dependence is read in
   instr_info * bb;
   Int bb_id;
   fun_entry * fe;

   //valid for the first instr of a BB
   Int cd_cnt;
   instr_info ** cds; //!=NULL when it is the first instruction in a BB
   instr_info * next_bb;
   instr_info * pred_ii;
   Int pred_id;


   Bool is_predicate;
   Bool true_propagate;
   Bool false_propagate;
   Addr jmp_pc;

   //source code info
   Char *   fn;
   Char *   file;
   Int      line;

   //global id
   Int      id;

   //run time
   Int cur_inst;
   Int lv_idx;

   //dyanmic information
   Int use_port;
   val_stream * uports[MAX_USE_PORT];
   UInt * visited;

   //forward slicing
   forward_edge ** edges;
   UInt * forward_visited;
   Int forward_inst;
#ifdef DEP_DISTANCE
   UInt * back_dist;
   UInt * forward_dist;
#endif

//#ifdef VALUE_WANTED
   Int allocated;
   UInt * values;
//#endif

//#ifdef TIMESTAMPED
   //only bb entry has timestamp and control flow info
   Int ts_alloced;
   UInt * tss;
   Addr * cfs;

//#endif


//#ifdef CERTIFIED
   UInt * certified;
//#endif
   Int input_related_alloc;
   UInt * input_related;
#ifdef MAPPING
   val_mapping * mapping;
#endif
   
//for relevant slicing, potential dependences, meaningful for loads
   Int pds_alloced;
   potential_dep ** pds;
   potential_cd * pcd;

};

/* data structure for dynamic switching */

//the set of load instances which potentially depends on the same predicate instance
typedef struct _switch_potential_edge_tail sw_potential_edge_tail;
struct _switch_potential_edge_tail {
     sw_potential_edge_tail * next;
     instr_info * use_ii; 
     Int inst;
     Int ts;
     UInt value;
};

typedef struct _switch_potential_edge_head sw_potential_edge_head;
struct _switch_potential_edge_head {
     sw_potential_edge_head * next;
     sw_potential_edge_head * hash_next;
     instr_info * pred_ii;
     Int inst;
     Int ts;
     sw_potential_edge_tail * tails;
};


typedef struct _io_log io_log;
struct _io_log {
  Int fid;
  Int file_pos;  //what file position the current buf is at
  Int count;
  Int offset;    //the current offset
  Char buf[IO_BUF_MAX];
};

typedef struct _slice_node slice_node;
struct _slice_node {
   slice_node * next;
   instr_info * ii;
   Int inst;
//#ifdef VALUE_WANTED
   UInt value;
   Int step;
//#endif

};

typedef struct _cd_fun_entry cd_fun_entry;
struct _cd_fun_entry {
   Addr entry_addr;
   Char fn[FN_LEN];
   fun_entry * fe;
};

typedef struct _error_node error_node;
struct _error_node {
    Char  file[FILE_LEN];
    Int line;
    Int inst;
};

// Declaration for the dependence buffer.
typedef struct _dep_buffer_element Dep_Buffer_Element;
struct _dep_buffer_element
{
   Bool isUsed;
   UInt defInstrID;
   UInt defInst;
   UInt useInstrID;
   UInt useInst;
};

// Data structures for slicing.
extern VgHashTable SliceDepTable;
extern VgHashTable ComputedSlice;

typedef struct _SliceEntrySorted SliceEntrySorted;
struct _SliceEntrySorted
{
   Addr instruction;
   UInt instance;
};

typedef struct _SliceEntry SliceEntry;
struct _SliceEntry // structure for a slice entry, which is stored in a hashtable
{
   SliceEntry *next;
   Addr instruction; // key
   UInt *instances; // list of instances for this instruction that are in the slice.
   UInt numInstances;
};

typedef struct _SliceInstanceInfo SliceInstanceInfo;
struct _SliceInstanceInfo // structure shows, for a given instruction instance, the set of other instruction instances it depends upon.
{
   Addr *instructions; // two parallel arrays
   UInt *instances;    // two parallel arrays
   UInt numEntries;
};

typedef struct _SliceInstruction SliceInstruction;
struct _SliceInstruction // structure designed to be placed into a hashtable (first 2 fields are important)
{
   SliceInstruction *next;
   Addr instruction; // key
   SliceInstanceInfo **instanceInfoList;
   UInt numInstances;
};

#endif
