
/*--------------------------------------------------------------------*/
/*--- Simple tool for counting UInstrs, using a C helper.          ---*/
/*---                                                    lk_main.c ---*/
/*--------------------------------------------------------------------*/

/*
   This file is part of Lackey, an example Valgrind tool that does
   some simple program measurement.

   Copyright (C) 2002-2004 Nicholas Nethercote
      njn25@cam.ac.uk

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.

   The GNU General Public License is contained in the file COPYING.
*/

#include "vg_skin.h"
#include "lk_helper.h"
#include "utility.h"
/* Nb: use ULongs because the numbers can get very big */
#define SHADOW1(tempreg) ((tempreg)+2)
#define VG_IS_FLAG_OVERLAP(set1,set2) \
   (( ((FlagSet)set1) & ((FlagSet)set2) ) != 0 )

typedef 
   enum {
      GETV = DUMMY_FINAL_UOPCODE + 1, 
      PUTV,
      INS  
   }
   MemCheckOpcode;

Addr helper_reg_dep_ports[7]={
    NULL,
    (Addr)&XZ_HELPER_reg_dep_port1,
    (Addr)&XZ_HELPER_reg_dep_port2,
    (Addr)&XZ_HELPER_reg_dep_port3,
    (Addr)&XZ_HELPER_reg_dep_port4,
    (Addr)&XZ_HELPER_reg_dep_port5,
    (Addr)&XZ_HELPER_reg_dep_port6};
#ifdef VALUE_WANTED
Addr helper_trace_values[5]={
     NULL,
    (Addr) & XZ_HELPER_trace_value1,
    (Addr) & XZ_HELPER_trace_value2,
    NULL,
    (Addr) & XZ_HELPER_trace_value4};
#endif
Int open_tid;
Int output_fp, output_fp_pos,output_fp_sim;
extern XZ_shadow_reg sh_regs[8];
extern VgHashTable instr_info_table;
extern VgHashTable fun_entry_table;

VgHashTable SliceDepTable;
VgHashTable ComputedSlice;

extern VG_(env_setenv) (Char***, Char*, Char*);

void SK_(pre_clo_init)(void)
{
   VG_(details_name)            ("Dynamic Dependence Graph");
   VG_(details_version)         (NULL);
   VG_(details_description)     ("an Valgrind tool");
   VG_(details_copyright_author)(
      "Copyright (C) 2005-2007, and GNU GPL'd, by Xiangyu Zhang.");
   VG_(details_bug_reports_to)  (VG_BUGS_TO);
   VG_(details_avg_translation_sizeB) ( 228 );

   VG_(needs_shadow_regs)         ();
   VG_(needs_extended_UCode)      ();
   VG_(needs_syscall_wrapper)     ();
   VG_(needs_shadow_memory)       ();

   open_tid=VG_(gettid)();
   //output_fp_sim= VG_(open)("SLICING.sim_output", VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,VKI_S_IRUSR|VKI_S_IWUSR);
   //output_fp= VG_(open)("SLICING.output", VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,VKI_S_IRUSR|VKI_S_IWUSR);
   //output_fp_pos= VG_(open)("SLICING.output_pos", VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,VKI_S_IRUSR|VKI_S_IWUSR);
   output_fp_sim= VG_(open)("/dev/null", VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,VKI_S_IRUSR|VKI_S_IWUSR);
   output_fp= VG_(open)("/dev/null", VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,VKI_S_IRUSR|VKI_S_IWUSR);
   output_fp_pos= VG_(open)("/dev/null", VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,VKI_S_IRUSR|VKI_S_IWUSR);
   if (output_fp < 0 || output_fp_pos < 0 || output_fp_sim<0) {
      VG_(message)(Vg_UserMsg, "FATAL: can't open output file");
      VG_(exit)(1);
   }

   VG_(XZ_shadow_num)=3;
   init_shadow_memory();
   instr_info_table=VG_(HT_construct)();
   fun_entry_table=VG_(HT_construct)();
   SliceDepTable = VG_(HT_construct)();
   ComputedSlice = VG_(HT_construct)();


   VG_(register_compact_helper) ((Addr) & XZ_HELPER_get_inst);
   VG_(register_compact_helper) ((Addr) & XZ_HELPER_increment_inst);
   VG_(register_compact_helper) ((Addr) & XZ_HELPER_setup_CD);
   VG_(register_compact_helper) ((Addr) & XZ_HELPER_reg_dep_port1);
   VG_(register_compact_helper) ((Addr) & XZ_HELPER_reg_dep_port2);
   VG_(register_compact_helper) ((Addr) & XZ_HELPER_load_mem_dep);
   VG_(register_compact_helper) ((Addr) & XZ_HELPER_store_mem_dep);
   VG_(register_compact_helper) ((Addr) & XZ_HELPER_get_timestamp);

#ifdef TIMESTAMPED 
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_trace_timestamp);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_trace_control_flow);
#endif

#ifdef VALUE_WANTED
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_trace_value1);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_trace_value2);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_trace_value4);
#endif
#ifdef TRACK_ARGV
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_track_argv);
#endif

#ifdef INTER_CD
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_create_dummy_CD);
#endif


   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_get);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_put);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_print_reg);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_terminate);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_BB);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_EBP_set_value);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_EBP_get_value);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_load_value);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_store_value);


   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_setup_CD_new);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_init_lvs_new);
   //VG_(register_noncompact_helper)((Addr) & XZ_HELPER_init_lvs);

   VG_(register_noncompact_helper) ((Addr) & XZ_HELPER_pred_dep);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_reg_dep_port3);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_reg_dep_port4);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_reg_dep_port5);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_reg_dep_port6);
/*
#define MUTT
#ifdef MUTT
   Char ** envp= VG_(client_envp);
   VG_(env_setenv) (&envp, "LD_LIBRARY_PATH", "/scratch/xyzhang/toolchain-i386-glibc/lib");
#endif
*/   
//#ifdef CONTROL_DEP
   read_cd_fun_entries("./cd/CONTROL_DEP");
//#endif
}

void SK_(post_clo_init)(void)
{
   // Initialize the dependence buffer to all empty structs
   initDepBuffer();
}

//Search for the predicate in a basic block(for control dependence)
static Int search_for_predicate(UCodeBlock * cb,int start_point)
{
           Int i;
           UInstr * instr;
           UInstr * cjmp=NULL;
           //for (i=VG_(get_num_instrs) (cb)-1;i>=0; i--) {
           for (i=start_point; i<VG_(get_num_instrs)(cb);i++) {
                  instr=VG_(get_instr) (cb,i);
                  //uFlagsRWU(cb, FlagsOSZACP, FlagsEmpty, FlagsEmpty)
                  //conditional branch
                  if (instr->opcode==JMP && instr->cond!=CondAlways&&instr->jmpkind!=JmpRet ) {
                          cjmp=instr;
                          break;
                  }
           }
           if (!cjmp) return -1;
           for (;i>=0;i--) {
                  instr=VG_(get_instr)(cb,i);
                  //if (instr->flags_w==FlagsOSZACP) {
                  if (VG_IS_FLAG_OVERLAP(instr->flags_w,cjmp->flags_r)) {
                          return i;
                  }
           }
           //print_cb(cb);
           //VG_(skin_panic)("Conditional branch found, however, predicate could not be found");
           return -1;
}

//set the ii and inst of the current instruction to the shadow register of any definition port
static __inline__ void XZ_set_instr_inst(UCodeBlock * cb, instr_info* ii, Int s1, Int s2)
{
   
     //uInstr0(cb,INS,0);
     uInstr2(cb, MOV,   4, Literal, 0, TempReg, s1);
     uLiteral(cb, (UInt) ii);

     
     //uInstr0(cb,INS,0);
     VG_(ccall_R_R) (cb, (Addr) & XZ_HELPER_get_inst, s1, s2, 1);
}


static void synth_XZ_getv (Int arch, Int reg_ii, Int reg_inst)
{
  Int REG;
  for (REG=R_EAX;REG<=R_EDI;REG++) {
      if (REG!=reg_ii && REG!=reg_inst) break;
  }
  VG_(emit_pushv_reg) (4,REG);
  VG_(emit_movv_lit_reg) (4, (UInt)sh_regs, REG);

  VG_(emit_movv_offregmem_reg)(4,8*arch,REG,reg_ii);
  VG_(emit_movv_offregmem_reg)(4,8*arch+4,REG,reg_inst);

  VG_(emit_popv_reg)(4,REG);

}

static void synth_XZ_putv (Int arch, Int reg_ii, Int reg_inst)
{
  Int REG;
  for (REG=R_EAX;REG<=R_EDI;REG++) {
      if (REG!=reg_ii && REG!=reg_inst) break;
  }

  VG_(emit_pushv_reg) (4,REG);
  VG_(emit_movv_lit_reg) (4, (UInt)sh_regs, REG);

  VG_(emit_movv_reg_offregmem)(4,reg_ii,8*arch,REG);
  VG_(emit_movv_reg_offregmem)(4,reg_inst,8*arch+4,REG);

  VG_(emit_popv_reg)(4,REG);
}

void SK_(emit_XUInstr) ( UInstr* u, RRegSet regs_live_before )
{
     switch(u->opcode) {
        case GETV:
            sk_assert(u->tag1==ArchReg&&u->tag2==RealReg && u->tag3==RealReg);
	    synth_XZ_getv(u->val1, u->val2, u->val3);
	    break;
        case PUTV:
            sk_assert(u->tag3==ArchReg&&u->tag1==RealReg && u->tag2==RealReg);
	    synth_XZ_putv(u->val3, u->val1, u->val2);
	    break;
    }
}

Bool SK_(sane_XUInstr)(Bool beforeRA, Bool beforeLiveness, UInstr* u)
{
#  define LIT0 (u->lit32 == 0)
#  define LITm (u->tag1 == Literal ? True : LIT0 )
#  define CC0 (u->flags_r == FlagsEmpty && u->flags_w == FlagsEmpty)
#  define SZi (u->size == 4 || u->size == 2 || u->size == 1)
#  define A1  (u->tag1 == ArchReg)
#  define A2  (u->tag2 == ArchReg)
#  define A3  (u->tag3 == ArchReg)
#  define TR1 (beforeRA ? (u->tag1 == TempReg) : (u->tag1 == RealReg))
#  define TR2 (beforeRA ? (u->tag2 == TempReg) : (u->tag2 == RealReg))
#  define TR3 (beforeRA ? (u->tag3 == TempReg) : (u->tag3 == RealReg))

#  define COND0    (u->cond         == 0)
#  define EXTRA4b0 (u->extra4b      == 0)
#  define SG_WD0   (u->signed_widen == 0)
#  define JMPKIND0 (u->jmpkind      == 0)
#  define CCALL0   (u->argc==0 && u->regparms_n==0 && u->has_ret_val==0 && \
                    ( beforeLiveness                                       \
                    ? u->regs_live_after == ALL_RREGS_LIVE                 \
                    : True ))
#  define XOTHER   (COND0 && EXTRA4b0 && SG_WD0 && JMPKIND0 && CCALL0)

    switch(u->opcode) {
	case GETV: return LIT0 && SZi && CC0 && A1 && TR2 && TR3 && XOTHER;
	case PUTV: return LIT0 && SZi && CC0 && TR1 && TR2 && A3 && XOTHER;
        case INS: return True;
   default:
      VG_(printf)("unhandled opcode: %u\n", u->opcode);
      VG_(skin_panic)("SK_(sane_XUInstr): unhandled opcode");
   }
}

Char* SK_(name_XUOpcode)(Opcode opc)
{
   switch (opc) {
      case GETV:    return "GETV";
      case PUTV:    return "PUTV";
      case INS: return "INS";
      default:
         VG_(printf)("unhandled opcode: %u\n", opc);
         VG_(skin_panic)("SK_(name_XUOpcode): unhandled case");
   }
}

void SK_(pp_XUInstr)(UInstr* u)
{
   switch (u->opcode) {
      case GETV: case PUTV:
         VG_(pp_UOperand)(u, 1, u->opcode==PUTV ? 4 : u->size, False);
         VG_(printf)(", ");
         VG_(pp_UOperand)(u, 2, u->opcode==GETV ? 4 : u->size, False);
         VG_(printf)(", ");
         VG_(pp_UOperand)(u, 3, u->opcode==GETV ? 4 : u->size, False);
         break;

      case INS: 
         break;
      default:
         VG_(printf)("unhandled opcode: %u\n", u->opcode);
         VG_(skin_panic)("SK_(pp_XUInstr): unhandled opcode");
   }

}


Int SK_(get_Xreg_usage)(UInstr* u, Tag tag, Int* regs, Bool* isWrites)
{
#  define RD(ono)    VG_UINSTR_READS_REG(ono, regs, isWrites)
#  define WR(ono)    VG_UINSTR_WRITES_REG(ono, regs, isWrites)

   Int n = 0;
   switch (u->opcode) {
      case GETV:    WR(2); WR(3);               break;
      case PUTV:    RD(1); RD(2);               break;
      case INS:                                break;

      default:
         VG_(printf)("unhandled opcode: %u\n", u->opcode);
         VG_(skin_panic)("SK_(get_Xreg_usage): unhandled opcode");
   }
   return n;

#  undef RD
#  undef WR
}


//To allocate space for CD, call when the function entry BB is instrumented.
static Int stack_alloc_space_find_index_new(UCodeBlock * cb_in, fun_entry * fe)
{
    Int i,j;
    UInstr * u_in, *tmp_in;
  
    for (i=0;i<VG_(get_num_instrs)(cb_in)-1;i++) {
       u_in = VG_(get_instr)(cb_in, i);
       if (u_in->opcode!=PUT) continue;
       if (u_in->val2==R_EBP) return i;// move %esp, %ebp 
    }//for
    return -1;
}

//mov %esp, %ebp
static void stack_alloc_space_instrument_new(UCodeBlock * cb, UInstr * u_in, fun_entry * fe)
{
                            //uInstr0(cb,INS,0);
                            uInstr2(cb, SUB, 4, Literal, 0, TempReg, u_in->val1);
                            //uLiteral(cb, 32); 
                            uLiteral(cb,fe->alloc_lv * PER_CD_SPACE+LAST_PREDICATE);
                            uInstr2(cb,PUT,4, TempReg, u_in->val1, ArchReg, R_ESP);
                            //uLiteral(cb,fe->alloc_lv * PER_CD_SPACE);

                            //VG_(copy_UInstr)(cb, u_in);

                            //ESP should be changed before initilization becuase ESP change will allocate the page, or else you could get segment fault.        
                            //init the allocated lvs to 0
                            //uInstr0(cb,INS,0);

                            //uInstr2(cb,GET, 4, ArchReg, R_EBP, TempReg, t_reg);

			    VG_(ccall_RL_0) (cb, (Addr) & XZ_HELPER_init_lvs_new, u_in->val1, (UInt)fe, 2);

}

static void stack_alloc_space_instrument_after_new (UCodeBlock * cb, UInstr * u_in, fun_entry * fe)
{
     uInstr2(cb, ADD, 4, Literal, 0, TempReg, u_in->val1);
     uLiteral(cb,fe->alloc_lv * PER_CD_SPACE+LAST_PREDICATE);
}

//find pop %ebp
static Int stack_free_space_find_index(UCodeBlock * cb_in)
{
     Int i,t;
     UInstr * u_in;
     t=-1;
   
     //find there is a return
     for (i=VG_(get_num_instrs) (cb_in)-1;i>=0;i--) {
        u_in=VG_(get_instr) (cb_in,i);
        if (u_in->opcode==JMP ) {
	/*get wrong for __int_malloc because the JMP is not necessary ret*/
        //if (u_in->opcode==JMP && u_in->jmpkind==JmpRet) {
	   t=i;
	   break;
	}
     }
     //VG_(printf) ("STEP 1: %d\n",t);
     if (t==-1) return -1;

     //find the PUT t, EBP
     for (i=t,t=-1;i>=0;i--) {
        u_in=VG_(get_instr) (cb_in,i);
	if (u_in->opcode==PUT && u_in->val2==R_EBP) {
	    t=i;
	    break;
        }
     }
     if (t==-1) return -1;
     //find PUT t, ESP
     for (i=t,t=-1;i>=0;i--) {
        u_in=VG_(get_instr) (cb_in,i);
	if (u_in->opcode==INCEIP) return -1;
	if (u_in->opcode==PUT && u_in->val2==R_ESP) {
	    t=i;
	    break;
        }
     }

     if (t==-1) return -1;

     //get the prvious LOAD (%esp), EBP 
     for (i=t,t=-1;i>=0;i--) {
        u_in=VG_(get_instr) (cb_in,i);
	if (u_in->opcode==INCEIP) return -1;
	if (u_in->opcode==LOAD) return i;
     }
     return -1;
}	   


//find pop %ebp
static Int stack_free_space_find_index_new(UCodeBlock * cb_in, Int * esp_reg)
{
     Int i,t;
     UInstr * u_in;
     t=-1;
     Int use;
   
     //find there is a return
     for (i=VG_(get_num_instrs) (cb_in)-1;i>=0;i--) {
        u_in=VG_(get_instr) (cb_in,i);
        if (u_in->opcode==JMP ) {
	/*get wrong for __int_malloc because the JMP is not necessary ret*/
        //if (u_in->opcode==JMP && u_in->jmpkind==JmpRet) {
	   t=i;
	   break;
	}
     }
     //VG_(printf) ("STEP 1: %d\n",t);
     if (t==-1) return -1;

     //find the PUT t, EBP
     for (i=t,t=-1;i>=0;i--) {
        u_in=VG_(get_instr) (cb_in,i);
	if (u_in->opcode==PUT && u_in->val2==R_EBP) {
	    use=u_in->val1;
            t=i;
	    break;
        }
     }

     if (t==-1) return -1;
     //find the defintion for 'use', normally LD (ts), td
     for (i=t,t=-1;i>=0;i--) {
        u_in=VG_(get_instr) (cb_in,i);
	if (u_in->opcode==INCEIP) return -1;
	if (u_in->val2==use) {
	    use=u_in->val1;
            t=i;
	    break;
        }
     }

     if (t==-1) return -1;
     //find the definition for 'use', normally ESP+x, use
     //get the prvious LOAD (%esp), EBP 
     for (i=t,t=-1;i>=0;i--) {
        u_in=VG_(get_instr) (cb_in,i);
	if (u_in->opcode==INCEIP) {
            *esp_reg=use;
	    return i;
        }
	if (u_in->val2==use) return i;
     }
     return -1;
}	   

static void stack_free_space_instrument_new(UCodeBlock * cb, Int reg, fun_entry * fe)
{
     uInstr2(cb, ADD, 4, Literal, 0, TempReg, reg);
     //uInstr2(cb, ADD, 4, Literal, 0, TempReg, u_get->val1);
     uLiteral(cb, fe->alloc_lv * PER_CD_SPACE+LAST_PREDICATE);
     //uInstr2(cb, PUT, 4, TempReg, u_get->val2, ArchReg, R_ESP);
}

//To allocate space for CD, call when the function entry BB is instrumented.
static Int stack_alloc_space_find_index(UCodeBlock * cb_in, fun_entry * fe)
{
    Int i,j;
    UInstr * u_in, *tmp_in;
    Bool start_counting=False;
    Int esp_def_cnt=0;
  
    for (i=0;i<VG_(get_num_instrs)(cb_in)-1;i++) {
       u_in = VG_(get_instr)(cb_in, i);
       if (u_in->opcode!=PUT) continue;
       if (u_in->val2==R_EBP) start_counting=True;
       if (!start_counting) continue;
       tmp_in = VG_(get_instr)(cb_in, i+1);
       if (u_in->val2==R_ESP){
           if (tmp_in->opcode==STORE) 	esp_def_cnt++;
	   else {//!=STORE
	        j=i-1;
                do {
		    tmp_in=VG_(get_instr)(cb_in,j);
	            if (tmp_in->opcode==SUB && tmp_in->tag1==Literal && tmp_in->val2==u_in->val1) {//the SUB instr 	
                        fe->slack=(esp_def_cnt-1) * 4;
			fe->orig_sp_offset=tmp_in->lit32+fe->slack;
		        return j;
                    }
		    j--;
		} while (j>=0);
                sk_assert(j>=0); //always failed at this point
            }
        }
     }//for
     return -1;
}

static void stack_alloc_space_instrument(UCodeBlock * cb, UInstr * u_in, fun_entry * fe,Int t_reg)
{
                            //uInstr0(cb,INS,0);
                            uInstr2(cb, SUB, 4, Literal, 0, TempReg, u_in->val1);
                            //uLiteral(cb, 32); 
                            uLiteral(cb,fe->alloc_lv * PER_CD_SPACE+LAST_PREDICATE);
                            //uLiteral(cb,fe->alloc_lv * PER_CD_SPACE);

                            VG_(copy_UInstr)(cb, u_in);

                            //ESP should be changed before initilization becuase ESP change will allocate the page, or else you could get segment fault.        
                            //init the allocated lvs to 0
                            //uInstr0(cb,INS,0);

                            uInstr2(cb,GET, 4, ArchReg, R_EBP, TempReg, t_reg);

			    VG_(ccall_RL_0) (cb, (Addr) & XZ_HELPER_init_lvs, t_reg, (UInt)fe, 2);

}


//it is called when ESP is the def port
static void stack_free_space_instrument(UCodeBlock * cb_in, UCodeBlock * cb, fun_entry * fe ,Int idx, UInstr * u_in)
{
    UInstr * tmp_in,*tmp_in1;
    Int j;

                   j=idx-1;
                   do {
                       tmp_in = VG_(get_instr)(cb_in, j);
                       tmp_in1 = VG_(get_instr)(cb_in, j-1);
                       if (tmp_in1->opcode==GET && tmp_in1->val1==R_ESP && tmp_in->opcode==ADD && tmp_in->tag1==Literal && tmp_in->val2==u_in->val1 && tmp_in->lit32==fe->orig_sp_offset-fe->slack ) {//so ugly hacking!!
                            //lv_finish=True;
                            //it is the substraction of the sp
                            uInstr0(cb,INS,0);
                            uInstr2(cb, ADD, 4, Literal, 0, TempReg, u_in->val1);
                            //uLiteral(cb, 32); 
                            uLiteral(cb,fe->alloc_lv * PER_CD_SPACE+LAST_PREDICATE);
                            return;
                       }
                       j--;
                    } while (j>=1);

                    //sk_assert (j>=1);
}
#ifdef TRACK_ARGV
static void instrument_main_entry(UCodeBlock * cb, Int  r_tmp, Int r_tmp1)
{
	  VG_(printf)("INSTRUMENT MAIN\n");
          uInstr2(cb, GET, 4, ArchReg, R_ESP, TempReg, r_tmp);
          uInstr2(cb, LEA1, 4, TempReg, r_tmp, TempReg, r_tmp1);
          uLiteral(cb, 0x4);
          uInstr2(cb, LOAD, 4, TempReg, r_tmp1, TempReg, r_tmp1);
          uInstr2(cb, LEA1, 4, TempReg, r_tmp, TempReg, r_tmp);
          uLiteral(cb,0x8);
          uInstr2(cb, LOAD, 4, TempReg, r_tmp, TempReg, r_tmp);
          VG_(ccall_RR_0) (cb, (Addr) & XZ_HELPER_track_argv, r_tmp1, r_tmp,2);
}
#endif 
 


static void instrument_predicate_with_ii(UCodeBlock * cb, UCodeBlock * cb_in, instr_info * bb, instr_info * ii, Int * r_tmp, Int * r_tmp1,Int * r_tmp2) 
{
          /*sk_assert(bb->lv_idx!=LV_NOT_ASSIGNED);*/      
          uInstr2(cb, GET, 4, ArchReg, R_EBP, TempReg, *r_tmp); 
	  //DEBUG
//#define DEBUG
//#undef DEBUG
#ifdef DEBUG
          VG_(ccall_RL_0) (cb, (Addr) & XZ_HELPER_debug_EBP_get_value, *r_tmp,0, 1);
#endif
          /*store instr_info into stack*/ 
          uInstr2(cb, LEA1, 4, TempReg, *r_tmp, TempReg, *r_tmp1);
          uLiteral(cb, (PER_CD_SPACE*bb->lv_idx+LAST_PREDICATE+CD_II_OFFSET-4));
	  uInstr2(cb, MOV, 4, Literal, 0, TempReg, * r_tmp2);
          uLiteral(cb, ii);
          uInstr2(cb, STORE, 4, TempReg,*r_tmp2, TempReg, *r_tmp1);
          //DEBUG
#ifdef DEBUG
          VG_(ccall_LRR_0)(cb, (Addr) & XZ_HELPER_debug_store_value, 0, *r_tmp1, ii,3);
#endif
          //store instance into stack
	  uInstr2(cb, LEA1, 4, TempReg, *r_tmp, TempReg, *r_tmp1);
          uLiteral(cb, (PER_CD_SPACE*bb->lv_idx+LAST_PREDICATE+CD_INST_OFFSET-4));

          VG_(ccall_L_R) (cb, (Addr) & XZ_HELPER_get_inst, ii, *r_tmp2, 1);
          uInstr2(cb, STORE, 4, TempReg, *r_tmp2, TempReg, *r_tmp1);
          //DEBUG
#ifdef DEBUG
          VG_(ccall_LRR_0)(cb, (Addr) & XZ_HELPER_debug_store_value, 0, *r_tmp1, *r_tmp2,3);
#endif
          //store timestamp into stack
	  uInstr2(cb, LEA1, 4, TempReg, *r_tmp, TempReg, *r_tmp1);
          uLiteral(cb, (PER_CD_SPACE*bb->lv_idx+LAST_PREDICATE+CD_TIMESTAMP_OFFSET-4));
          VG_(ccall_R_R) (cb, (Addr) &XZ_HELPER_get_timestamp, *r_tmp, *r_tmp,1);
          uInstr2(cb, STORE, 4, TempReg, *r_tmp, TempReg, *r_tmp1);
          //DEBUG
#ifdef DEBUG
          VG_(ccall_LRR_0)(cb, (Addr) & XZ_HELPER_debug_store_value, 0, *r_tmp1, *r_tmp ,3);
#endif
#undef DEBUG
}




static void instrument_predicate (UCodeBlock * cb, UCodeBlock * cb_in, instr_info * bb, UInstr * u_in, Int * r_tmp, Int * r_tmp1) 
{
          Int r_dest;
 
          /*find the destination reg*/ 
	  if (u_in->tag3==TempReg) { 
	      r_dest=u_in->val3; 
	  } else if (u_in->tag2==TempReg) {
	      r_dest=u_in->val2;
	  } else if (u_in->tag1==TempReg) {
	      r_dest=u_in->val1;
          } else {
	      //print_cb(cb_in);
	      //VG_(skin_panic) ("Could not find the definition port.");
	      return;
          } 
          /*sk_assert(bb->lv_idx!=LV_NOT_ASSIGNED);*/      
          uInstr2(cb, GET, 4, ArchReg, R_EBP, TempReg, *r_tmp); 
	  //DEBUG
//#define DEBUG
//#undef DEBUG
#ifdef DEBUG
          VG_(ccall_RL_0) (cb, (Addr) & XZ_HELPER_debug_EBP_get_value, *r_tmp,0, 1);
#endif
          /*store instr_info into stack*/ 
          uInstr2(cb, LEA1, 4, TempReg, *r_tmp, TempReg, *r_tmp1);
          uLiteral(cb, (PER_CD_SPACE*bb->lv_idx+LAST_PREDICATE+CD_II_OFFSET-4));
          uInstr2(cb, STORE, 4, TempReg, SHADOW(r_dest), TempReg, *r_tmp1);
          //DEBUG
#ifdef DEBUG
          VG_(ccall_LRR_0)(cb, (Addr) & XZ_HELPER_debug_store_value, 0, *r_tmp1, SHADOW(r_dest),3);
#endif
          //store instance into stack
	  uInstr2(cb, LEA1, 4, TempReg, *r_tmp, TempReg, *r_tmp1);
          uLiteral(cb, (PER_CD_SPACE*bb->lv_idx+LAST_PREDICATE+CD_INST_OFFSET-4));
          uInstr2(cb, STORE, 4, TempReg, SHADOW1(r_dest), TempReg, *r_tmp1);
          //DEBUG
#ifdef DEBUG
          VG_(ccall_LRR_0)(cb, (Addr) & XZ_HELPER_debug_store_value, 0, *r_tmp1, SHADOW1(r_dest),3);
#endif
          //store timestamp into stack
	  uInstr2(cb, LEA1, 4, TempReg, *r_tmp, TempReg, *r_tmp1);
          uLiteral(cb, (PER_CD_SPACE*bb->lv_idx+LAST_PREDICATE+CD_TIMESTAMP_OFFSET-4));
          VG_(ccall_R_R) (cb, (Addr) &XZ_HELPER_get_timestamp, *r_tmp, *r_tmp,1);
          uInstr2(cb, STORE, 4, TempReg, *r_tmp, TempReg, *r_tmp1);
          //DEBUG
#ifdef DEBUG
          VG_(ccall_LRR_0)(cb, (Addr) & XZ_HELPER_debug_store_value, 0, *r_tmp1, *r_tmp ,3);
#endif
#undef DEBUG
}


/* Multiple ucode could correspond to one x86 code*/
#define TRACK_INSTR \
	consider_defined=True; \
	if (!ii) { \
            get_bb_info(cur_x86_addr,&cur_fileCC, &cur_fnCC, &line); \
	    ii=get_instr_info(cur_x86_addr,fe,line,INVALID_UPORT,&is_new_instr,fe->cd_available); \
  	    /*VG_(printf) ("ii=%x, ii->bb=%x, bb=%x\n",ii->orig_addr, ii->bb, bb);*/ \
	    if (fe->cd_available && ii->bb!=bb) { \
		bb=ii->bb; \
                predicate_idx=search_for_predicate(cb_in,i); \
	    } \
        }

#define UPDATE_INSTR_UPORT \
        if (ii && ii->use_port==INVALID_UPORT) { \
	   ii->use_port=cur_uport; \
	}
  

#define SETUP_CD  \
         if (fe->cd_available && i>stack_init_idx && !stack_freed) { \
  	    /*VG_(printf) ("CD ii->bb=%x, bb=%x\n", ii->bb, bb);*/ \
            uInstr2(cb, GET, 4, ArchReg, R_EBP, TempReg, r_tmp); \
	    VG_(ccall_LR_0) (cb, (Addr)&XZ_HELPER_setup_CD_new, (UInt)ii, r_tmp,2); \
         }

#define INCREMENT_COUNTER \
	    VG_(ccall_L_0) (cb, (Addr) & XZ_HELPER_increment_inst, (UInt) ii, 1);
//MOVE the shadow of rs to rd depends on tag
#define INSTRUMENT_OPERAND_SHADOW(tag, rs, rd) \
     switch (tag) {  \
	case TempReg: \
	    uInstr2(cb, MOV, 4, TempReg, SHADOW(rs),TempReg, SHADOW(rd)); \
	    uInstr2(cb, MOV, 4, TempReg, SHADOW1(rs), TempReg, SHADOW1(rd)); \
	    break; \
        case Literal: \
	    TRACK_INSTR \
            XZ_set_instr_inst(cb,ii,SHADOW(rd), SHADOW1(rd)); \
	    break; \
	case ArchReg: \
	    uInstr3(cb, GETV, 4, ArchReg, rs, TempReg, SHADOW(rd), TempReg, SHADOW1(rd)); \
	    break; \
      }
   
#define INSTRUMENT_DEBUG_TERMINATE \
      VG_(ccall_0_0) (cb, (Addr) & XZ_HELPER_debug_terminate);

#define INSTRUMENT_DEBUG_GET\
      VG_(ccall_LL_0) (cb, (Addr) & XZ_HELPER_debug_get, i, u_in->val1,2);

#define INSTRUMENT_DEBUG_PUT\
      VG_(ccall_LRR_0) (cb, (Addr) & XZ_HELPER_debug_put, i, SHADOW(u_in->val1),SHADOW1(u_in->val1), 3);

#define INSTRUMENT_DEBUG \
      VG_(ccall_L_0) (cb, (Addr) & XZ_HELPER_debug, i, 1);

#define REG_DEP(use_reg) \
    internal_defined=False;  \
    /*VG_(printf) ("XXXXXXXXX %d CUR_UPORT=%d\n",i, cur_uport);*/ \
    for (k=0;k<cur_def;k++) {  \
	if (use_reg == internal_defs[k]) {  \
             internal_defined=True;  \
	     break;             \
        } \
    } \
    if (!internal_defined) {  /*defined by previous x86 instruction*/ \
	 /*VG_(printf) ("\tXXXXXXXXX %d CUR_UPORT=%d\n",i, cur_uport);*/ \
         VG_(ccall_LRR_0) (cb, helper_reg_dep_ports[cur_uport++], (UInt)ii, SHADOW(use_reg), SHADOW1(use_reg),3); \
    }

#define SETUP_TWO_EDGES \
	    REG_DEP(u_in->val1) \
	    REG_DEP(u_in->val2) \
	    XZ_set_instr_inst(cb,ii,SHADOW(u_in->val2), SHADOW1(u_in->val2)); 


#define MEM_DEP(addr_reg) \
	   VG_(ccall_LLR_0) (cb, (Addr) &XZ_HELPER_load_mem_dep, (UInt)ii, cur_uport++, addr_reg, 3);

/*
#define DEFINED_BY_EBP(reg) \
          ebp_used=False; \
	  reg_defed=False; \
	  for (j=i-1;j>=0;j--) { \
	      tmp_in=VG_(get_instr)(cb_in,j); \
              k=VG_(get_reg_usage) ( tmp_in, TempReg, instr_regs, instr_regs_iswrite); \
	      for (l=0;l<k; l++) { \
		 if (instr_regs_iswrite[l] && instr_regs[l]==reg) { \
		     reg_defed=True; \
                 } \
              }\
	      if (reg_defed) { \
		  if (tmp_in->val1==R_EBP || tmp_in->val2==R_EBP || tmp_in->val3==R_EBP) \
			ebp_used=True; \
		  break; \
              } \
          }
      
        
		   
#define OFFSET_EBP_ACCESS_LEA1 \
	  if (fe->cd_available) \
		DEFINED_BY_EBP(u_in->val1) \
          if (ebp_used  && ((Int) u_in->lit32)>0) { \
		uLiteral(cb, ((Int)u_in->lit32)+fe->alloc_lv*PER_CD_SPACE+LAST_PREDICATE); \
          }

#define OFFSET_EBP_ACCESS_LEA2 \
	  if (fe->cd_available) { \
		DEFINED_BY_EBP(u_in->val1) \
	        if (!ebp_used)  \
		    DEFINED_BY_EBP(u_in->val2) \
          } \
          if (ebp_used  && ((Int) u_in->lit32)>0) { \
		uLiteral(cb, ((Int)u_in->lit32)+fe->alloc_lv*PER_CD_SPACE+LAST_PREDICATE); \
          }
*/

#define OFFSET_EBP_ACCESS_LEA1 \
	  /* After the function entry BB*/ \
        /*	VG_(printf)("TRY OFFSET LEA1 %x ebp_reg=%d\n",cur_x86_addr,ebp_reg);*/ \ 
	  if (ebp_reg==u_in->val1  && ((Int) u_in->lit32)>0) { \
		/*VG_(printf)("OFFSET LEA1 %x\n",cur_x86_addr);*/ \
		uLiteral(cb, ((Int)u_in->lit32)+fe->alloc_lv*PER_CD_SPACE+LAST_PREDICATE); \
          }
	
#define OFFSET_EBP_ACCESS_LEA2 \
	  /* After the function entry BB*/ \
	  if ((ebp_reg==u_in->val1 || ebp_reg==u_in->val2) && ((Int) u_in->lit32)>0) { \
		/*VG_(printf)("OFFSET LEA2\n");*/ \
		uLiteral(cb, ((Int)u_in->lit32)+fe->alloc_lv*PER_CD_SPACE+LAST_PREDICATE); \
          }
	
#define MAX_INTERNAL_DEF 20

Int main_entry=0;
UCodeBlock* SK_(instrument)(UCodeBlock* cb_in, Addr orig_addr)
{
   UCodeBlock* cb;
   Int         i,j,k,l, def;
   UInstr*     u_in, * tmp_in;
   Char        fnname[100];
   Int line;
   fileCC * cur_fileCC;
   fnCC * cur_fnCC;
   fun_entry * fe;
   instr_info * ii=NULL, *cmp_ii=NULL;
   Bool is_fun_entry;
   Addr cur_x86_addr;
   Bool is_new_instr;
   Int r_tmp, r_dest, r_tmp1,r_tmp2;
   
   Int stack_init_idx=-1;
   Int predicate_idx=-1, pred_idx=-1;
   Int cur_uport=1;//0 is for control dependence
   Int internal_defs[MAX_INTERNAL_DEF];
   Int cur_def=0;
   Bool internal_defined;
   Int instr_regs[VG_MAX_REGS_USED];
   Bool instr_regs_iswrite[VG_MAX_REGS_USED];

   Bool stack_freed=False;
   instr_info * bb=NULL,*pred_ii=NULL,*start_bb;
   Bool after_get_ebp=False;
   Int stack_free_idx=-1;
   Int ebp_reg=-1;
   Int instr_reg_used;
   Int esp_reg=-1;
   Bool consider_defined; 
   Bool debug_terminate=False,jmpStamped=False;
   Int last_define=-1, last_def_size=0;
   FlagSet cmp_flags;

   //if (1) return cb_in;
   cb = VG_(setup_UCodeBlock)(cb_in);
   r_tmp=newTemp(cb);
   r_tmp1=SHADOW(r_tmp);
   r_tmp2=SHADOW1(r_tmp);
   //get debug information
   get_bb_info(orig_addr,&cur_fileCC, &cur_fnCC, &line);
#ifdef TRACK_ARGV
   VG_(printf) ("CUR_FN=%s\n",cur_fnCC->fn);
   if (!main_entry && !VG_(strcmp)(cur_fnCC->fn,"main")) {
	main_entry=1;
        instrument_main_entry (cb,r_tmp,r_tmp1);
   }
#endif
   //if it is a new function, the control dependence is read in and the instructions for bb entries will be created with INVALID line and uport
   fe=get_fun_entry_new(orig_addr, cur_fileCC, cur_fnCC);

   //if (!VG_(strcmp)("_Exit",fe->fn->fn)) return cb_in;
   //fe->cd_available=False;


   if (fe->cd_available) bb=search_for_bb(orig_addr, fe);

   //VG_(printf)("search_for bb done\n"); 

   /* Count call to dlrr(), if this BB is dlrr()'s entry point */
   is_fun_entry=VG_(get_fnname_if_entry)(orig_addr, fnname, 100);

   //pred_idx is for tracing predicate results
   pred_idx=search_for_predicate(cb_in,0);

   if (fe->cd_available) {
	predicate_idx=pred_idx;
	//VG_(printf) ("XXXXXXX predicate_idx=%d\n",predicate_idx);
   }
   if (fe->cd_available) stack_free_idx=stack_free_space_find_index_new(cb_in,&esp_reg);
   //VG_(printf)("stack free space done\n"); 
   if (is_fun_entry && fe->cd_available) stack_init_idx=stack_alloc_space_find_index_new(cb_in, fe);
   //VG_(printf)("stack alloc space done\n"); 
   cur_x86_addr=orig_addr;
   ii=NULL;
   ebp_reg=-1;
   //DEBUG
   //VG_(ccall_L_0) (cb, (Addr) & XZ_HELPER_debug_BB,orig_addr,1);
   for (i = 0; i < VG_(get_num_instrs)(cb_in); i++) {
      consider_defined=False;
      //ii=NULL;
      //if (cur_x86_addr==0x805320a) debug_terminate=True;
      u_in = VG_(get_instr)(cb_in, i);
      //VG_(pp_UInstr) ( i, u_in );
      
      
      if (i==stack_init_idx) {
	  //VG_(printf)("XXXXX stack_init_idx=%d\n",i);
	  stack_alloc_space_instrument_new(cb,u_in,fe);
	  ebp_reg=u_in->val1;
      }
   /* 
      if (i==stack_free_idx) {
	  VG_(printf)("XXXXX stack_free_idx=%d\n",i);
          stack_free_space_instrument_new(cb, u_in,fe);
	  stack_freed=True;
      }
*/
      //INSTRUMENT_DEBUG
      switch (u_in->opcode) {
         case GETSEG:
            sk_assert(u_in->tag2 == TempReg);
	    TRACK_INSTR
            XZ_set_instr_inst(cb,ii,SHADOW(u_in->val2), SHADOW1(u_in->val2));
            VG_(copy_UInstr)(cb, u_in);
	    break;

         case INCEIP:
	    if (ii) {
		SETUP_CD //this instruction has been tracked, thus CD need to establish
		UPDATE_INSTR_UPORT //update the number of use ports for current instruction
		INCREMENT_COUNTER
            }

#ifdef TIMESTAMPED
	    //TRACING control flow
	    if (bb && cur_x86_addr==bb->orig_addr) { //trace timestamp only when it is the first ii for a bb
		if (!ii) { //ii is not set up for dependence
		    //INCREMENT_COUNTER
	            VG_(ccall_L_0) (cb, (Addr) & XZ_HELPER_increment_inst, (UInt) bb, 1);
		}
	        VG_(ccall_L_0) (cb, (Addr) &XZ_HELPER_trace_timestamp, (UInt)bb, 1);
	    }
#endif


	    cur_x86_addr+=u_in->val1;

            if (fe->cd_available) bb=search_for_bb(cur_x86_addr, fe);
#ifdef VALUE_WANTED
	    if (ii&&last_define!=-1) {
		//VG_(printf) ("TRACE VALUE @ %d, size=%d\n",last_define, last_def_size);
	        VG_(ccall_LR_0) (cb, helper_trace_values[last_def_size], (UInt)ii, last_define,2);
	        //VG_(ccall_LR_0) (cb, (Addr) &XZ_HELPER_trace_value, (UInt)ii, last_define,2);
            }
	    last_define=-1;
	    last_def_size=0;
#endif
            VG_(copy_UInstr)(cb, u_in);
            ii=NULL;
	    cur_uport=1;
	    cur_def=0;
	    after_get_ebp=False;
            break;

         case LOAD:	
	    TRACK_INSTR
	    REG_DEP (u_in->val1)
            MEM_DEP (u_in->val1)

            XZ_set_instr_inst(cb,ii,SHADOW(u_in->val2), SHADOW1(u_in->val2));
            VG_(copy_UInstr)(cb, u_in);
            //DEBUG
//#define DEBUG
#ifdef DEBUG
            VG_(ccall_RR_0) (cb, (Addr) & XZ_HELPER_debug_load_value, u_in->val1, u_in->val2, 2);
#endif
	    break;
	   
         case STORE:
	    TRACK_INSTR
	    REG_DEP(u_in->val1)
	    REG_DEP(u_in->val2)

	    VG_(ccall_LLR_0) (cb, (Addr) &XZ_HELPER_store_mem_dep, (UInt)ii,u_in->size, u_in->val2,3);
	    //VG_(ccall_LR_0) (cb, (Addr) &XZ_HELPER_store_mem_dep, (UInt)ii,u_in->size, u_in->val2,2);
            //DEBUG
#ifdef DEBUG            
	    VG_(ccall_LRR_0)(cb,(Addr) &XZ_HELPER_debug_store_value, cur_x86_addr, u_in->val2, u_in->val1,3);
#endif
            VG_(copy_UInstr)(cb, u_in);
            break;
         //they are not tracked
	 case GET:
            uInstr3(cb, GETV,4, ArchReg, u_in->val1,
				TempReg, SHADOW(u_in->val2),
				TempReg, SHADOW1(u_in->val2));
	    //DEBUG
            //INSTRUMENT_DEBUG_GET
            VG_(copy_UInstr)(cb,u_in);
            //VG_(ccall_LLR_0)(cb, (Addr) & XZ_HELPER_debug_print_reg,cur_x86_addr,u_in->val2,u_in->val2,3); 
	    if (u_in->val1==R_EBP) {
	       //DEBUG
               //VG_(ccall_RL_0) (cb, (Addr) & XZ_HELPER_debug_EBP_get_value, u_in->val2,orig_addr, 1);
            }
	    if (fe->cd_available && u_in->val1==R_EBP) { 
		//after_get_ebp=True;
		ebp_reg=u_in->val2;
            }
	    break;

	 case PUT:
            uInstr3(cb, PUTV,4, TempReg, SHADOW(u_in->val1),
				TempReg, SHADOW1(u_in->val1),
				ArchReg,u_in->val2);
	    //INSTRUMENT_DEBUG_PUT
            //VG_(ccall_LLR_0)(cb, (Addr) & XZ_HELPER_debug_print_reg,cur_x86_addr,u_in->val1,u_in->val1,3); 
	    
            //is it the end of a function, restore ESP
            /*
	    if (fe->cd_available && u_in->val2==R_ESP && !stack_freed) {
		stack_freed=True;
	        stack_free_space_instrument(cb_in, cb, fe, i, u_in);	
	    }
            */	   
            VG_(copy_UInstr)(cb,u_in);
	    if (u_in->val2==R_EBP) {
               //DEBUG
               //VG_(ccall_RL_0) (cb, (Addr) & XZ_HELPER_debug_EBP_set_value, u_in->val1, orig_addr,1);
            }

	    break;

         case MOV:
	    INSTRUMENT_OPERAND_SHADOW(u_in->tag1, u_in->val1, u_in->val2)
            VG_(copy_UInstr)(cb,u_in);
	    break;

	 case LEA1:
	    INSTRUMENT_OPERAND_SHADOW(TempReg, u_in->val1, u_in->val2)
            VG_(copy_UInstr)(cb,u_in);
            OFFSET_EBP_ACCESS_LEA1
	    break;

	 case LEA2:
	    TRACK_INSTR
	    REG_DEP(u_in->val1)
	    REG_DEP(u_in->val2)

            XZ_set_instr_inst(cb,ii,SHADOW(u_in->val3), SHADOW1(u_in->val3));
            VG_(copy_UInstr)(cb,u_in);
            OFFSET_EBP_ACCESS_LEA2
	    break;
/*
	 case RCL: case RCR: case SHL:
         case SHR: case SAR:
         case ROL: case ROR:
	    INSTRUMENT_OPERAND_SHADOW(u_in->tag1, u_in->val1, u_in->val2)
            VG_(copy_UInstr)(cb,u_in);
	    break;
*/

         case CC2VAL:
	 if (VG_IS_FLAG_OVERLAP(u_in->flags_r, cmp_flags)) {
	    TRACK_INSTR
	    ii->is_predicate=True;
	    VG_(ccall_LLL_0)(cb, (Addr) & XZ_HELPER_pred_dep, ii,cur_uport, cmp_ii,3);
	    cur_uport++;
            XZ_set_instr_inst(cb,ii,SHADOW(u_in->val1), SHADOW1(u_in->val1));
         } else {
            for (j=i-1;j>=0;j--) {
                  tmp_in=VG_(get_instr)(cb_in,j);
                  if (VG_IS_FLAG_OVERLAP(u_in->flags_r,tmp_in->flags_w)) {
                  //if (tmp_in->flags_w==FlagsOSZACP) {
                       break;
                  }
            }

            if (j>=0) {//found the comparison
                        //r_cd is the result temp reg to store the result bdd
                 if (tmp_in->tag3==TempReg) {
		      INSTRUMENT_OPERAND_SHADOW(TempReg, tmp_in->val3, u_in->val1)
                 } else if (tmp_in->tag2==TempReg) {
		      INSTRUMENT_OPERAND_SHADOW(TempReg, tmp_in->val2, u_in->val1)
                 } else if (tmp_in->tag1==TempReg) {
		      INSTRUMENT_OPERAND_SHADOW(TempReg, tmp_in->val1, u_in->val1)
                 } else {
	              TRACK_INSTR
                      XZ_set_instr_inst(cb,ii,SHADOW(u_in->val1), SHADOW1(u_in->val1));
		 }
            } else {
	        TRACK_INSTR
                XZ_set_instr_inst(cb,ii,SHADOW(u_in->val1), SHADOW1(u_in->val1));
            }
         }
            VG_(copy_UInstr)(cb, u_in);
            break;

	 case CMOV:
	    uInstr2(cb, CMOV, 4, TempReg, SHADOW(u_in->val1), TempReg, SHADOW(u_in->val2));
	    uInstr2(cb, CMOV, 4, TempReg, SHADOW1(u_in->val1), TempReg, SHADOW1(u_in->val2));
            VG_(copy_UInstr)(cb, u_in);
            break;
	 
         case RCL: case RCR: case SHL:
         case SHR: case SAR:
         case ROL: case ROR:

         case ADD: case SUB:
         case ADC: case SBB:
         case MUL: case XOR:
	    switch (u_in->tag1) {  
 	       case TempReg:
		    TRACK_INSTR
	            //SETUP_TWO_EDGES
                    REG_DEP(u_in->val1);
                    REG_DEP(u_in->val2);
                    XZ_set_instr_inst(cb,ii,SHADOW(u_in->val2), SHADOW1(u_in->val2));
		    break; 
	       case Literal: 
	            break; 
	       case ArchReg: 
		    TRACK_INSTR
                    uInstr3(cb, GETV, 4, ArchReg, u_in->val1, TempReg, SHADOW(r_tmp), TempReg, SHADOW1(r_tmp)); 
		    REG_DEP(r_tmp);
		    REG_DEP(u_in->val2);

                    XZ_set_instr_inst(cb,ii,SHADOW(u_in->val2), SHADOW1(u_in->val2));
	            //SETUP_TWO_EDGES
	            break; 
            }
            VG_(copy_UInstr)(cb, u_in);
            break;

         case AND: case OR:
	    TRACK_INSTR
            REG_DEP(u_in->val1);
            REG_DEP(u_in->val2);
            XZ_set_instr_inst(cb,ii,SHADOW(u_in->val2), SHADOW1(u_in->val2));
	    //SETUP_TWO_EDGES
            VG_(copy_UInstr)(cb, u_in);
            break;

         case CALLM:
	    TRACK_INSTR
            j=i-1;
	    tmp_in=VG_(get_instr) (cb_in,j);
            while (tmp_in->opcode!=CALLM_S) {
	        if (tmp_in->opcode == PUSH) {
		    REG_DEP(tmp_in->val1)
		    if(cur_uport>10) VG_(exit)(0);
		}
		//VG_(printf) ("i=%d, j=%d, tmp_in->opcode=%d, CALLM_S=%d\n",i, j, tmp_in->opcode, CALLM_S);
		j--;
                tmp_in=VG_(get_instr) (cb_in,j);
            }
	    sk_assert(cur_uport<MAX_USE_PORT);

            VG_(copy_UInstr)(cb,u_in);
            j=i+1;
	    tmp_in=VG_(get_instr) (cb_in,j);
            while (tmp_in->opcode!=CALLM_E) {
	        if (tmp_in->opcode == POP) {
                     XZ_set_instr_inst(cb,ii,SHADOW(tmp_in->val1), SHADOW1(tmp_in->val1));
                }
                j++;
                tmp_in=VG_(get_instr) (cb_in,j);
            }

	    break;

         case JMP:
#ifdef INTER_CD
	    if (u_in->jmpkind==JmpCall) {
		if (!ii) {
	           TRACK_INSTR
		}
                uInstr2(cb, GET, 4, ArchReg, R_ESP, TempReg, r_tmp); 
	        VG_(ccall_LR_0) (cb, (Addr) &XZ_HELPER_create_dummy_CD, ii, r_tmp, 2);
		//ii->cur_uport++;
	    }
#endif

//#ifdef TRACING_PREDICATE_VAL
	    if (pred_ii) {//if there is comparison, need to establish ii for jump and trace the value of jump
	        if (!ii) {
	            TRACK_INSTR
                }

	        if (u_in->cond!=CondAlways) {	
		    ii->is_predicate=True;
		   //VG_(printf)("JMP tracing %x, pred_ii=%x\n",ii->orig_addr,pred_ii->orig_addr);
		    //establish the dependence between comparison and jump
		    VG_(ccall_LLL_0)(cb, (Addr) & XZ_HELPER_pred_dep, ii,cur_uport, pred_ii,3);
		    cur_uport++;
		    
		    if (predicate_idx!=-1) {
                       if (bb->lv_idx==LV_NOT_ASSIGNED) {	
                             //VG_(printf) ("NOT ASSIGNED bb is %x, bb->bb= %x\n",bb->orig_addr, bb->bb->orig_addr);
                             //goto CD_OUT;	     
                       } else {
                             instrument_predicate_with_ii(cb, cb_in, bb, ii, &r_tmp, & r_tmp1, &r_tmp2);	  
		       }
		    }

                    SETUP_CD
	            UPDATE_INSTR_UPORT
                    INCREMENT_COUNTER

#ifdef VALUE_WANTED
		    VG_(ccall_LL_0) (cb, (Addr) & XZ_HELPER_trace_value1, ii,1,2); 
#endif

		} else { //CondAlways
		    //Jcondition
		    //instrumentation for tracing 0
		    //jmp
#ifdef VALUE_WANTED
		    VG_(ccall_LL_0) (cb, (Addr) & XZ_HELPER_trace_value1, ii,0,2); 
#endif
		}

	    } else  if (ii) {//otherwise , it is switch case statement / Jmp $EAX or Jmp-C
               SETUP_CD
	       UPDATE_INSTR_UPORT
               INCREMENT_COUNTER
            }

            //instrument for switch case
            if (fe->cd_available && predicate_idx==-1 && bb->lv_idx!=LV_NOT_ASSIGNED && u_in->jmpkind!=JmpRet) {//for switch case?

		//VG_(printf) ("WARNING, no predicate, but it is a CD\n");
                instrument_predicate(cb,cb_in,bb, u_in, &r_tmp, &r_tmp1);
            } 

#ifdef TIMESTAMPED
	    if (bb && u_in->jmpkind!=JmpCall && u_in->jmpkind!=JmpRet) {
		//basic block has only one instruction, no INCEIP
		if (cur_x86_addr==bb->orig_addr && !jmpStamped) {
		     jmpStamped=True;
 		     if (!ii) { //ii is not set up for dependence
		         //INCREMENT_COUNTER
	                 VG_(ccall_L_0) (cb, (Addr) & XZ_HELPER_increment_inst, (UInt) bb, 1);
		     }
	             VG_(ccall_L_0) (cb, (Addr) &XZ_HELPER_trace_timestamp, (UInt)bb, 1);
		}

		//the JCondition will be overwritten by Juncondition
		if (u_in->tag1==Literal) {
		   VG_(ccall_LL_0) (cb, (Addr) & XZ_HELPER_trace_control_flow, bb, u_in->lit32,2);
		} else {
		   VG_(ccall_LR_0) (cb, (Addr) & XZ_HELPER_trace_control_flow, bb, u_in->val1,2);
	        }
	    }
#endif

            VG_(copy_UInstr)(cb, u_in);
            ii=NULL;
	    cur_uport=1;
	    cur_def=0;
	    if (u_in->cond!=CondAlways &&bb&& bb->pred_ii) {
		bb->pred_ii->jmp_pc=cur_x86_addr;
 	    }
            break;
	   
	    
         default:
            VG_(copy_UInstr)(cb, u_in);
            break;
      }

/*
      for (k=0;k<instr_reg_used;k++) {
	  if (instr_regs_iswrite[k]) {
	     //ebp_reg is redefined
	     if (ebp_reg==instr_regs[k]) ebp_reg=-1;
	     internal_defs[cur_def++]=instr_regs[k];
          }
      }
*/

//#ifdef TRACING_PREDICATE_VAL
      if (i==pred_idx && ii) {
	  //ii->is_predicate=True;
	  pred_ii=ii;
	  if (predicate_idx!=-1) bb->pred_ii=ii;
      }
     
      if (u_in->flags_w && ii) {
	 cmp_ii=ii;
	 cmp_flags=u_in->flags_w;
      }
//#endif

/*
      if (i==predicate_idx) {
	  //VG_(printf) ("i=%d, predicate_idx=%d\n",i,predicate_idx);
	  sk_assert(bb!=NULL);
	  //Bool need_to_increment=False;
	  if (!ii) {
	      TRACK_INSTR
          }

	  //SETUP_CD
	  //UPDATE_INSTR_UPORT
          //need_to_increment=True;
          
	  bb->pred_ii=ii;
          if (bb->lv_idx==LV_NOT_ASSIGNED) {	
             VG_(printf) ("NOT ASSIGNED bb is %x, bb->bb= %x\n",bb->orig_addr, bb->bb->orig_addr);
             goto CD_OUT;	     
          }
          instrument_predicate(cb, cb_in, bb, u_in, &r_tmp, & r_tmp1);	  
   
      
      }//predicate	                 
*/


      /* 
      if (debug_terminate) {
          INSTRUMENT_DEBUG_TERMINATE
      }
      */
CD_OUT:
      //if (u_in->opcode!=GET) {
          instr_reg_used=VG_(get_reg_usage) ( u_in, TempReg, instr_regs, instr_regs_iswrite);

          for (k=0;k<instr_reg_used;k++) {
	     if (instr_regs_iswrite[k]) {
	        //ebp_reg is redefined
	        if (u_in->opcode!=GET && ebp_reg==instr_regs[k]) ebp_reg=-1;
                if (consider_defined) {
	            internal_defs[cur_def++]=instr_regs[k];
		}
		//trace value on last define
		last_define=instr_regs[k];
	  	last_def_size=u_in->size;
             }
          }
      //}
      //pop %ebp
      if (i==stack_free_idx) {
	  //G_(printf)("XXXXX stack_free_idx=%d\n",i);
          if (u_in->opcode==INCEIP) {
             stack_free_space_instrument_new(cb, esp_reg,fe);
          } else {
	     stack_free_space_instrument_new(cb, u_in->val2, fe);
          }
	  stack_freed=True;
      }

//#define ENABLE_DEBUG   
#ifdef ENABLE_DEBUG
      if (cur_x86_addr==0x80531D5  //000000
          || cur_x86_addr==0x805319b   //000000
         ) {
            instr_reg_used=VG_(get_reg_usage) (u_in, TempReg, instr_regs, instr_regs_iswrite);
            for (k=0;k<instr_reg_used;k++) {
               if (instr_regs_iswrite[k]) {
                   VG_(ccall_LLR_0)(cb, (Addr) & XZ_HELPER_debug_print_reg,cur_x86_addr,instr_regs[k],instr_regs[k],3); 
               } 
	    }//for
       }
#endif
   }
#ifdef ENABLE_DEBUG
   if (orig_addr==0x804829d || orig_addr==0x80481f0) {
       VG_(printf)("PRINT BB %x\n", orig_addr);
       print_cb(cb);
   }
#endif

      // print_cb(cb);
   VG_(free_UCodeBlock)(cb_in);
   return cb;
}

void SK_(fini)(Int exitcode)
{
  Int i, j;
  Addr startInstruction = 0x0;
  Int startInstance = -1;
  Int pid=VG_(gettid)();
  Char buf[128];
  Char str[128];
  Bool result;
  
  VG_(printf)("Finishing for %d...\n",pid);

  // Perform slicing if a slicing starting point was specified.
  Char * startPoint = VG_(getSliceStartPoint)();
  if(startPoint[0] != NULL)
  {
     // Get the instruction and instance number from which to start slicing.
     Bool okay = False;
     for(i = 0; i < strlen(startPoint); i++)
     {
        if(startPoint[i] == '_')
        {
           okay = True;
           break;
        }
     }
     if(okay)
     {
        memset(&str, '\0', 128);
        if(strlen(startPoint) >= 2 && startPoint[0] == '0' && startPoint[1] == 'x')
        {
           for(i = 0; startPoint[i] != '_'; i++)
           {
              str[i] = startPoint[i];
           }
           sscanf(&str, "%x", &startInstruction);
           i++;
           memset(&str, '\0', 128);
           j = 0;
           for(; i < strlen(startPoint); i++)
           {
              str[j] = startPoint[i];
              j++;
           }
           sscanf(&str, "%d", &startInstance);
        }
        else
        {
           str[0] = '0';
           str[1] = 'x';
           for(i = 0; startPoint[i] != '_'; i++)
           {
              str[i+2] = startPoint[i];
           }
           sscanf(&str, "%x", &startInstruction);
           i++;
           memset(&str, '\0', 128);
           j = 0;
           for(; i < strlen(startPoint); i++)
           {
              str[j] = startPoint[i];
              j++;
           }
           sscanf(&str, "%d", &startInstance);
        }
        
        
        
        if(startInstruction == 0x0)
        {
           VG_(printf)("Could not slice: invalid slice starting instruction (0x%x)\n", startInstruction);
        }
        else if(startInstance < 0)
        {
           VG_(printf)("Could not slice: invalid slice starting instance number (%d)\n", startInstance);
        }
        else
        {
           VG_(printf)("Slicing from instr 0x%x, instance %d... ", startInstruction, startInstance);
           
           // Slice from the specified starting point, depending upon which format of tracing information
           // was collected.
           if(VG_(doOnlyBufferOutput)() == True)
           {
              result = sliceLimited(startInstruction, startInstance);
           }
           else
           {
              result = sliceComprehensive(startInstruction, startInstance);
           }
           
           if(result == True)
           {
              VG_(printf)("written to file \"SLICE.txt\"\n");
           }
        }
     }
     else
     {
        VG_(printf)("Could not slice: invalid slice starting point specified (requires '_')\n");
     }
  }
  
  // Output the tracing information as long as the --noout flag was not specified.
  if(!(VG_(doNoOutput)()))
  {
     if(VG_(doOnlyBufferOutput)() == True)
     {
        VG_(printf)("Writing LIMITED trace to file... ");
        VG_(sprintf)(buf,"TRACE.txt");
        write_buffer_trace(buf);
        VG_(printf)("written to file \"TRACE.txt\"\n");
     }
     else
     {
        VG_(printf)("Writing COMPREHENSIVE trace to file... ");
        VG_(sprintf)(buf,"TRACE.txt");
        write_dependence(buf);
        VG_(printf)("written to file \"TRACE.txt\"\n");
     }
  }

  //VG_(sprintf)(buf,"SLICING.%d.instr_dic",pid);
  //VG_(printf) ("finishing...\n");
  //write_instr_dic(buf);
  //VG_(printf) ("finishing... instr_dic done\n");
  //VG_(sprintf)(buf,"SLICING.%d.pdg",pid);
  //write_dependence(buf);
  //VG_(printf) ("finishing... pdg done\n");
  //if (pid==open_tid) {
     //VG_(printf)("closing output files for %d...\n",pid);
     //VG_(close)(output_fp);
     //VG_(close)(output_fp_pos);
  //}
  
  VG_(printf)("All done!\n");
}
//___NR_write
#define SYSCALL_READ 3
#define SYSCALL_WRITE 4
#define SYSCALL_EXECVE 11

void * SK_(pre_syscall) (ThreadId tid, UInt syscallno, Bool isBlocking)
{
  Int fd, i,j,k;
  Char * buf;
  Char ** argv;
  size_t count;
  void * tst;
  static Int total_count=0;
  Char t_buf[1024];
  Char c_buf[1024];
  instr_info * ii;
  Int inst;
  Char c;

  //if (0) {
  Int pid=VG_(gettid)();
  if (syscallno==SYSCALL_WRITE) {

        //VG_(printf) ("XXXXX SYS_WRITE (tid=%d):\n",pid);
        tst=VG_(XZ_get_threadstate_3_arg)(tid, &fd, &buf, &count);
        //VG_(printf) ("\tXXXXX SYS_WRITE good 1 count=%d,fd=%d, buf=%x\n",count,fd,buf);
        //VG_(printf) ("\tXXXXX SYS_WRITE buf[0]=%c, buf[15]=%c, count=%d\n",buf[0],buf[15],count);
	if (buf) {

	   //VG_(printf) ("SYS_WRITE:%s\n",buf);
	   //VG_(printf) ("\tCOUNT=%d\n",count);
	   VG_(sprintf) (t_buf, "(PID:%d) from %d, count=%d\n",pid,total_count,count);
	   //VG_(write) (output_fp, (void*) t_buf, VG_(strlen)(t_buf));
	   //VG_(write) (output_fp, (void*) buf,count);
	   VG_(write) (output_fp_pos, (void*) t_buf, VG_(strlen)(t_buf));
	   VG_(write) (output_fp_sim, (void*) t_buf, VG_(strlen)(t_buf));
	                //VG_(printf) ("OUTPUT %s\n",buf);
/*
           for (i=0;i<count;i++) {
	       c=buf[i];
	       VG_(sprintf)(t_buf,"%c",c);
               VG_(write)(output_fp_pos,(void*) t_buf, VG_(strlen)(t_buf));
           }
*/
           k=0;  
           for (i=0;i<count;i++) {
               ii=(instr_info*) get_ii_vbytes4_ALIGNED((Addr)(buf+i));
               inst=get_inst_vbytes4_ALIGNED((Addr)(buf+i));
	       c=buf[i];
	       
	       if (!c) c='0';
	       if ((UInt)ii!=0xffffffff && ii ) {
		   VG_(sprintf) (t_buf,"&* %d,%d *&'%c'\n",ii->id,inst,c);
	           VG_(write) (output_fp_sim, (void*) t_buf, VG_(strlen)(t_buf));
		   VG_(sprintf) (t_buf,"%c(%d,%d)",c,ii->id,inst);
	       } else {
		   VG_(sprintf) (t_buf,"%c",c);

               }
	       //VG_(printf)("%s\n",t_buf);
	       VG_(write) (output_fp_pos, (void*) t_buf, VG_(strlen)(t_buf));
	       c_buf[k++]=c;
	       if (c=='\n') {
		   c_buf[k]='\0';
		   VG_(write) (output_fp_sim, (void*) c_buf, VG_(strlen)(c_buf));
		   k=0;
	       }
           }
	   
	   total_count+=count;
	   //VG_(printf)("SYS_WRITE done\n");
        }
  } else if (syscallno==SYSCALL_EXECVE) {
        //VG_(printf) ("XXXXX SYS_EXECVE (tid=%d):\n",pid);
        tst=VG_(XZ_get_threadstate_exec_argv)(tid, &argv);
      
        //VG_(printf) ("\tXXXXX SYS_WRITE good 1 count=%d,fd=%d, buf=%x\n",count,fd,buf);
        //VG_(printf) ("\tXXXXX SYS_WRITE buf[0]=%c\n",buf[0]);
	if (argv) {
	   //VG_(printf) ("SYS_WRITE:%s\n",buf);
	   //VG_(printf) ("\tCOUNT=EXECVE\n",count);
	   VG_(sprintf) (t_buf, "(PID:%d) from %d, count=%d\n",pid,total_count,count);
	   VG_(write) (output_fp, (void*) t_buf, VG_(strlen)(t_buf));
	   VG_(write) (output_fp, (void*) buf,count);
	   VG_(write) (output_fp_pos, (void*) t_buf, VG_(strlen)(t_buf));
	   VG_(write) (output_fp_sim, (void*) t_buf, VG_(strlen)(t_buf));
	                //VG_(printf) ("OUTPUT %s\n",buf);
           i=0;
           while (argv[i]) {
	      buf=argv[i];
              for (j=0;j<VG_(strlen)(argv[i]);j++) {
                  ii=(instr_info*) get_ii_vbytes4_ALIGNED((Addr)(buf+j));
                  inst=get_inst_vbytes4_ALIGNED((Addr)(buf+j));
	          if ((UInt)ii!=0xffffffff && ii) {
		     VG_(sprintf) (t_buf,"&* %d,%d *&'%c'\n",ii->id,inst,buf[j]);
	             VG_(write) (output_fp_sim, (void*) t_buf, VG_(strlen)(t_buf));
		     VG_(sprintf) (t_buf,"%c(%d,%d)",buf[j],ii->id,inst);
	          } else {
		     VG_(sprintf) (t_buf,"%c",buf[j]);
                  }
	          VG_(write) (output_fp_pos, (void*) t_buf, VG_(strlen)(t_buf));
	      }//for
	      i++;
           }
	   
	   total_count+=count;
	   //VG_(printf)("SYS_WRITE done\n");
        }
/*
        i=0;
	while (argv[i]) {
	   VG_(printf) ("argv[%d]=%s\n",i,argv[i]);
	   i++;
        }
*/
  }
  //VG_(printf) ("XXXXX SYS_WRITE (tid=%d) DONE:\n",pid);
  return (void*) True;
}

__inline__ void  SK_(post_syscall) ( ThreadId tid, UInt syscallno,
                           void* pre_result, Int res, Bool isBlocking )
{
#ifdef TRACK_INPUT
  Int fd, i,j;
  Char * buf;
  Char ** argv;
  size_t count;
  void * tst;
  static Int total_count=0;
  Char t_buf[1024];
  instr_info * ii;
  Int inst;
  Char c;

  //if (0) {
  Int pid=VG_(gettid)();
  if (syscallno==SYSCALL_READ) {

        VG_(printf) ("XXXXX SYS_READ (tid=%d):\n",pid);
        tst=VG_(XZ_get_threadstate_3_arg)(tid, &fd, &buf, &count);
        //VG_(printf) ("\tXXXXX SYS_WRITE good 1 count=%d,fd=%d, buf=%x\n",count,fd,buf);
        //VG_(printf) ("\tXXXXX SYS_WRITE buf[0]=%c, buf[15]=%c\n",buf[0],buf[15]);
	if (buf&& res) {
/*
	   VG_(printf) ("SYS_WRITE:%s\n",buf);
	   VG_(printf) ("\tCOUNT=%d\n",count);
	   VG_(sprintf) (t_buf, "(PID:%d) from %d, count=%d\n",pid,total_count,count);
*/
	   VG_(printf) ("READ(%d)\n",res);
	   //VG_(printf) ("READ(%d)=%s\n",res,buf);
           for (i=0;i<res;i++) {
               set_ii_vbytes4_ALIGNED((Addr)(buf+i),INPUT_RELEVANT);
	       VG_(printf) ("set '%c' at %x\n",buf[i], buf+i);
               set_inst_vbytes4_ALIGNED((Addr)(buf+i),buf[i]);
           }
	 }
   }
	   //VG_(write) (outp
     //VG_(printf) ("POST good\n");
#endif
}


VG_DETERMINE_INTERFACE_VERSION(SK_(pre_clo_init), 2/1)
//VG_DETERMINE_INTERFACE_VERSION(SK_(pre_clo_init), 4./8)


/*--------------------------------------------------------------------*/
/*--- end                                                lk_main.c ---*/
/*--------------------------------------------------------------------*/

